# isort: skip_file

from dotenv import load_dotenv
from typing import Union, List, Any
import logging

__version__ = "0.1.1"
load_dotenv()


# Ignore import sorting, since LLMEngine needs to be imported first
from grazier.engines.llm import LLMEngine  # noqa: E402
from grazier.engines.chat import Conversation, LLMChat, Speaker, ConversationTurn  # noqa: F401, E402
from grazier.engines.image import ILMEngine  # noqa: E402


def get(name: str, chat: bool = False, type: str = "llm", **kwargs: Any) -> Union[LLMChat, LLMEngine]:
    """Get an engine by name."""
    name = name.lower()

    if chat:
        # Deprecated
        logging.warning("The `chat` argument is deprecated. Please use `type='chat'` instead.")
        type = "chat"

    if type == "chat":
        if name in LLMChat.list_models():
            return LLMChat.from_string(name, **kwargs)
        else:
            raise ValueError(f"Chat engine {name} not found")
    elif type == "image":
        if name in ILMEngine.list_models():
            return ILMEngine.from_string(name, **kwargs)
        else:
            raise ValueError(f"Image engine {name} not found")
    elif type == "llm":
        if name in LLMEngine.list_models():
            return LLMEngine.from_string(name, **kwargs)
        else:
            raise ValueError(f"LLM engine {name} not found")

    raise ValueError(f"Invalid engine type {type} (Select from 'chat', 'image', 'llm')")


def list_models(chat: bool = False, type: str = "llm") -> List[str]:
    """List available LLMs."""
    if chat:
        # Deprecated
        logging.warning("The `chat` argument is deprecated. Please use `type='chat'` instead.")
        type = "chat"

    if type == "chat":
        return LLMChat.list_models()
    elif type == "image":
        return ILMEngine.list_models()
    elif type == "llm":
        return LLMEngine.list_models()

    raise ValueError(f"Invalid engine type {type} (Select from 'chat', 'image', 'llm')")



import logging
from typing import Any, List, Optional

from transformers import AutoModelForCausalLM, AutoTokenizer, TFAutoModelForCausalLM
from transformers.pipelines import PIPELINE_REGISTRY, pipeline

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.huggingface import check_huggingface_model_files_are_local
from grazier.utils.python import singleton

from .instruct_pipeline import InstructionTextGenerationPipeline

PIPELINE_REGISTRY.register_pipeline(
    task="dolly-instruct",
    pipeline_class=InstructionTextGenerationPipeline,
    pt_model=AutoModelForCausalLM,
    tf_model=TFAutoModelForCausalLM,
    type="text",
)


class DollyEngine(LLMChat):
    def __init__(
        self,
        model: str,
        device: Optional[str] = None,
    ) -> None:
        super().__init__(device=device)

        self._pipeline = pipeline(
            "dolly-instruct",
            model=model,
            tokenizer=AutoTokenizer.from_pretrained(model),
            framework="pt",
            device=self.device,
            trust_remote_code=True,
        )

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        # If the length of the conversation is not a single user turn, warn
        system_turn_list = [c for c in conversation if c.speaker == Speaker.SYSTEM]
        if len(system_turn_list) > 1:
            logging.warning(
                "DollyEngine is designed to be used with a single user (and optionally, a single system) turn. This may not be the case."
            )
        if len(system_turn_list) > 0:
            system_turn = system_turn_list[-1]
        else:
            system_turn = None

        non_system_turns = [c for c in conversation if c.speaker != Speaker.SYSTEM]
        if len(non_system_turns) != 1 or non_system_turns[0].speaker != Speaker.USER:
            logging.warning(
                "DollyEngine is designed to be used with a single user (and optionally, a single system) turn. This may not be the case."
            )

        input_string = [c for c in non_system_turns if c.speaker == Speaker.USER][-1].text
        outputs = self._pipeline(
            input_string,
            do_sample=n_completions > 1,
            num_return_sequences=n_completions,
            intro_blurb=system_turn.text if system_turn is not None else None,
            **kwargs,
        )

        return [ConversationTurn(text=o["generated_text"], speaker=Speaker.AI) for o in outputs]  # type: ignore


@register_engine
@singleton
class DollyV23B(DollyEngine):
    name = ("Dolly v2 3B", "dolly-v2-3b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="databricks/dolly-v2-3b",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("databricks/dolly-v2-3b")


@register_engine
@singleton
class DollyV27B(DollyEngine):
    name = ("Dolly v2 7B", "dolly-v2-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="databricks/dolly-v2-7b",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("databricks/dolly-v2-7b")


@register_engine
@singleton
class DollyV212B(DollyEngine):
    name = ("Dolly v2 12B", "dolly-v2-12b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="databricks/dolly-v2-12b",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("databricks/dolly-v2-12b")


@register_engine
@singleton
class DollyV16B(DollyEngine):
    name = ("Dolly v1 6B", "dolly-v1-6b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="databricks/dolly-v1-6b",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("databricks/dolly-v1-6b")


@register_engine
@singleton
class MPT7BInstruct(DollyEngine):
    name = ("MPT 7B Instruct", "mpt-7b-instruct")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="mosaicml/mpt-7b-instruct",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("mosaicml/mpt-7b-instruct")


@register_engine
@singleton
class MPT30BInstruct(DollyEngine):
    name = ("MPT 30B Instruct", "mpt-30b-instruct")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__(
            model="mosaicml/mpt-30b-instruct",
            device=device,
        )

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("mosaicml/mpt-30b-instruct")



from abc import ABC, abstractmethod
from typing import Tuple


class Engine(ABC):
    @property
    @abstractmethod
    def name(self) -> Tuple[str, str]:
        """Returns a tuple of (Pretty Name, CLI name) of the language model."""
        raise NotImplementedError()

    @staticmethod
    def configure(*args, **kwargs) -> None:
        """Configures the engine."""
        raise NotImplementedError("This engine does not support automated configuration.")

    @staticmethod
    def is_configured() -> bool:
        """Returns True if the engine is configured configuration, False otherwise."""
        return "Unknown"

    @staticmethod
    def requires_configuration() -> bool:
        """Returns True if the engine is requires configuration, False otherwise."""
        return True



import logging
import random

import pytest

from grazier.engines.llm import LLMEngine


@pytest.mark.parametrize(
    "engine",
    [
        "llama-7B",
        "llama-13B",
        "llama-30B",
        "llama-60B",
    ],
)
def test_huggingface_llm_engine(engine: str) -> None:
    if engine in (
        "llama-13B",
        "llama-30B",
        "llama-60B",
    ):
        pytest.skip("Model too large for CI")

    _engine = LLMEngine.from_string(engine)
    random_number = random.randint(0, 100)
    response = _engine(f"My name, followed by a colon with the number {random_number} is:")

    for r in response:
        assert len(r.strip()) > 0, f"Response is empty: {r}"
        if str(random_number) not in r:
            logging.warning(f'Number "{random_number}" not found in response "{r}"')



import glob
import os
from typing import Any, List, Optional

from transformers import LlamaForCausalLM, LlamaTokenizer

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.python import singleton


class HuggingFaceLlamaLMEngine(LLMEngine):
    def __init__(
        self,
        model: str,
        weight_root: str,
        max_new_tokens: int = 128,
        device: Optional[str] = None,
    ) -> None:
        super().__init__(device=device)
        self._max_new_tokens = max_new_tokens
        self.tokenizer = LlamaTokenizer.from_pretrained(f"{os.environ.get(weight_root, '')}{model}")
        self._generator = LlamaForCausalLM.from_pretrained(
            f"{os.environ.get(weight_root, '')}{model}", device_map="auto"
        )

    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        input = self.tokenizer(prompt, return_tensors="pt").input_ids.to(self._generator.device)

        # Handle the kwargs
        _params = {
            "max_new_tokens": kwargs.get("max_new_tokens", kwargs.pop("max_tokens", self._max_new_tokens)),
            "num_return_sequences": n_completions,
            "temperature": kwargs.get("temperature", None),
        } | kwargs

        outputs = self._generator.generate(
            input,
            do_sample=n_completions > 1,
            **_params,
        )

        # Strip the prompt from the output
        outputs = outputs[:, input.shape[-1] :]
        # Decode and return the output
        outputs = self.tokenizer.batch_decode(outputs, skip_special_tokens=True, clean_up_tokenization_spaces=False)

        return outputs

    @staticmethod
    def _is_configured(weight_root: str, model: str) -> bool:
        # Check to see if the weights are available in the weight_root
        if weight_root is None:
            return False
        if not os.path.exists(os.path.join(weight_root, model)):
            return False

        # Check to see if the files are available:
        files = [
            "config.json",
            "pytorch_model.bin.index.json",
            "tokenizer_config.json",
            "generation_config.json",
            "special_tokens_map.json",
            "tokenizer.model",
        ]

        # Check to see if at least one model shard is available
        # pytorch_model-{n_layers + 1}-of-{n_layers + 1}.bin
        shards = list(glob.glob(os.path.join(weight_root, model, "pytorch_model-*.bin")))
        if len(shards) == 0:
            return False

        for file in files:
            if not os.path.exists(os.path.join(weight_root, model, file)):
                return False

        return True


@register_engine
@singleton
class Llama7B(HuggingFaceLlamaLMEngine):
    name = ("LLAMA 7B", "llama-7B")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("7B", "LLAMA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return HuggingFaceLlamaLMEngine._is_configured(os.environ.get("LLAMA_WEIGHTS_ROOT", "."), "7B")


@register_engine
@singleton
class Llama13B(HuggingFaceLlamaLMEngine):
    name = ("LLAMA 13B", "llama-13B")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("13B", "LLAMA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return HuggingFaceLlamaLMEngine._is_configured(os.environ.get("LLAMA_WEIGHTS_ROOT", "."), "13B")


@register_engine
@singleton
class Llama30B(HuggingFaceLlamaLMEngine):
    name = ("LLAMA 30B", "llama-30B")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("30B", "LLAMA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return HuggingFaceLlamaLMEngine._is_configured(os.environ.get("LLAMA_WEIGHTS_ROOT", "."), "30B")


@register_engine
@singleton
class Llama65B(HuggingFaceLlamaLMEngine):
    name = ("LLAMA 65B", "llama-65B")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("65B", "LLAMA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return HuggingFaceLlamaLMEngine._is_configured(os.environ.get("LLAMA_WEIGHTS_ROOT", "."), "65B")



import logging

import pytest

from grazier.engines.chat import Conversation, LLMChat, Speaker


@pytest.mark.parametrize("engine", ["dolly-v2-3b", "mpt-7b-instruct"])
def test_dolly_llm_engine(engine: str) -> None:
    # Construct a conversation
    conversation = Conversation()
    conversation.add_turn("You are an intelligent AI named Jason.", speaker=Speaker.SYSTEM)
    conversation.add_turn("Your name, followed by a colon with the number 42 is:", speaker=Speaker.USER)

    _engine = LLMChat.from_string(engine)
    responses = _engine(conversation)
    for r in responses:
        assert r.text.strip() != ""
        if "Jason" not in r.text:
            logging.warning(f'Name "Jason" not found in response "{r.text}"')
        if "42" not in r.text:
            logging.warning(f'Number "42" not found in response "{r.text}"')


@pytest.mark.parametrize("engine", ["dolly-v2-3b", "mpt-7b-instruct"])
def test_dolly_chat_engine(engine: str) -> None:
    # Construct a conversation
    conversation = Conversation()
    conversation.add_turn("You are an intelligent AI named Jason.", speaker=Speaker.SYSTEM)
    conversation.add_turn("Your name, followed by a colon with the number 42 is:", speaker=Speaker.USER)

    _engine = LLMChat.from_string(engine)
    responses = _engine(conversation)
    for r in responses:
        assert r.text.strip() != ""
        if "Jason" not in r.text:
            logging.warning(f'Name "Jason" not found in response "{r.text}"')
        if "42" not in r.text:
            logging.warning(f'Number "42" not found in response "{r.text}"')



import glob
import logging
import os
from typing import Any, List, Optional, Tuple

from transformers import LlamaForCausalLM, LlamaTokenizer

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import singleton


class LlamaLMEngine(LLMChat):
    def __init__(
        self,
        model: str,
        weight_root: str,
        max_new_tokens: int = 128,
        device: Optional[str] = None,
    ) -> None:
        super().__init__(device=device)
        self._max_new_tokens = max_new_tokens
        self.tokenizer = LlamaTokenizer.from_pretrained(f"{os.environ.get(weight_root, '')}{model}")
        self._generator = LlamaForCausalLM.from_pretrained(
            f"{os.environ.get(weight_root, '')}{model}", device_map="auto"
        )

    def _build_prompt_from_conversation(self, conversation: Conversation) -> Tuple[str, bool]:
        raise NotImplementedError()

    def _extract_last_turn(self, outputs: List[str], last_prompt_is_user: bool) -> List[str]:
        raise NotImplementedError()

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        prompt, last_prompt_is_user = self._build_prompt_from_conversation(conversation)

        input = self.tokenizer(prompt, return_tensors="pt").input_ids.to(self._generator.device)

        # Handle the kwargs
        _params = {
            "max_new_tokens": kwargs.get("max_new_tokens", kwargs.pop("max_tokens", self._max_new_tokens)),
            "num_return_sequences": n_completions,
            "temperature": kwargs.get("temperature", None),
        } | kwargs

        outputs = self._generator.generate(
            input,
            do_sample=n_completions > 1,
            **_params,
        )

        # Strip the prompt from the output
        outputs_filtered = outputs[:, input.shape[-1] :]
        # Decode and return the output
        outputs_filtered = self.tokenizer.batch_decode(
            outputs_filtered, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )

        # Pick out the last continued turn
        # NOTE: This has been removed, since we really do want to return everything.
        # outputs = self._extract_last_turn(outputs_filtered, last_prompt_is_user)

        return [
            ConversationTurn(text=o, speaker=Speaker.AI if last_prompt_is_user else Speaker.USER)
            for o in outputs_filtered
        ]

    @staticmethod
    def _is_configured(weight_root: Optional[str], model: str) -> bool:
        # Check to see if the weights are available in the weight_root
        if weight_root is None:
            return False
        if not os.path.exists(os.path.join(weight_root, model)):
            return False

        # Check to see if the files are available:
        files = [
            "config.json",
            "pytorch_model.bin.index.json",
            "tokenizer_config.json",
            "generation_config.json",
            "special_tokens_map.json",
            "tokenizer.model",
        ]

        # Check to see if at least one model shard is available
        # pytorch_model-{n_layers + 1}-of-{n_layers + 1}.bin
        shards = list(glob.glob(os.path.join(weight_root, model, "pytorch_model-*.bin")))
        if len(shards) == 0:
            return False

        for file in files:
            if not os.path.exists(os.path.join(weight_root, model, file)):
                return False

        return True


class KoalaLMEngine(LlamaLMEngine):
    def _build_prompt_from_conversation(self, conversation: Conversation) -> Tuple[str, bool]:
        # Step 1: Build the prompt from the conversation
        prompt = "BEGINNING OF CONVERSATION: "
        for turn in conversation.turns:
            if turn.speaker in (Speaker.USER, Speaker.SYSTEM):
                prompt += "USER: "
                prompt += turn.text + " "
            elif turn.speaker == Speaker.AI:
                prompt += "GPT: "
                prompt += turn.text + " "
                prompt += "</s>"

        # Append the next prompt
        last_prompt_is_user = True
        if conversation.turns[-1].speaker in (Speaker.USER, Speaker.SYSTEM):
            prompt += "GPT: "
            last_prompt_is_user = True
        else:
            logging.warning(
                "The last turn in the conversation is from the AI, however this model does not (usually) support continuing conversation from the user perspective."
            )
            prompt += "USER: "
            last_prompt_is_user = False

        return prompt, last_prompt_is_user

    def _extract_last_turn(self, outputs: List[str], last_prompt_is_user: bool) -> List[str]:
        return [o.split("GPT:" if last_prompt_is_user else "USER:")[0].strip() for o in outputs]


class Vicuna11LMEngine(LlamaLMEngine):
    def _build_prompt_from_conversation(self, conversation: Conversation) -> Tuple[str, bool]:
        # Step 1: Build the prompt from the conversation
        prompt = ""
        for idx, turn in enumerate(conversation.turns):
            if turn.speaker == Speaker.SYSTEM:
                if idx == 0:
                    prompt += f"{turn.text} "
                else:
                    logging.warning(
                        "Tried to add a system turn to the conversation at index != 0. This is not supported by the Vicuna 1.1 style model."
                    )

            if turn.speaker == Speaker.USER:
                prompt += "USER: "
                prompt += turn.text + " "
            elif turn.speaker == Speaker.AI:
                prompt += "ASSISTANT: "
                prompt += turn.text + " "
                prompt += "</s>"

        # Append the next prompt
        last_prompt_is_user = True
        if conversation.turns[-1].speaker in (Speaker.USER, Speaker.SYSTEM):
            prompt += "ASSISTANT: "
            last_prompt_is_user = True
        else:
            logging.warning(
                "The last turn in the conversation is from the AI, however this model does not (usually) support continuing conversation from the user perspective."
            )
            prompt += "USER: "
            last_prompt_is_user = False

        return prompt, last_prompt_is_user

    def _extract_last_turn(self, outputs: List[str], last_prompt_is_user: bool) -> List[str]:
        return [o.split("ASSISTANT:" if last_prompt_is_user else "USER:")[0].strip() for o in outputs]


class AlpacaLMEngine(LlamaLMEngine):
    def _build_prompt_from_conversation(self, conversation: Conversation) -> Tuple[str, bool]:
        # Step 1: Build the prompt from the conversation
        prompt = ""
        for idx, turn in enumerate(conversation.turns):
            if turn.speaker == Speaker.SYSTEM:
                if idx == 0:
                    prompt += f"{turn.text} "
                else:
                    logging.warning(
                        "Tried to add a system turn to the conversation at index != 0. This is not supported an Alpaca style model."
                    )

            if turn.speaker == Speaker.USER:
                prompt += "### Instruction:\n\n "
                prompt += turn.text + " "
            elif turn.speaker == Speaker.AI:
                prompt += "### Response:\n\n "
                prompt += turn.text + " "
                prompt += "</s>"

        # Append the next prompt
        last_prompt_is_user = True
        if conversation.turns[-1].speaker in (Speaker.USER, Speaker.SYSTEM):
            prompt += "### Response:\n\n "
            last_prompt_is_user = True
        else:
            logging.warning(
                "The last turn in the conversation is from the AI, however this model does not (usually) support continuing conversation from the user perspective."
            )
            prompt += "### Instruction:\n\n "
            last_prompt_is_user = False

        return prompt, last_prompt_is_user

    def _extract_last_turn(self, outputs: List[str], last_prompt_is_user: bool) -> List[str]:
        return [o.split("### Response:" if last_prompt_is_user else "### Instruction:")[0].strip() for o in outputs]


class AllenAILlamaLMEngine(LlamaLMEngine):
    def _build_prompt_from_conversation(self, conversation: Conversation) -> Tuple[str, bool]:
        # Step 1: Build the prompt from the conversation
        prompt = ""
        for idx, turn in enumerate(conversation.turns):
            if idx == 0 and turn.speaker == Speaker.SYSTEM:
                prompt += f"{turn.text}\n\n"
            elif idx == 0:
                prompt += "Below is an instruction that describes a task. Write a response that appropriately completes the request.\n\n"
            elif turn.speaker == Speaker.SYSTEM:
                logging.warning(
                    "Tried to add a system turn to the conversation at index != 0. This is not supported an Alpaca style model."
                )

            if turn.speaker == Speaker.USER:
                prompt += "<|user|>\n"
                prompt += turn.text + "\n"
            elif turn.speaker == Speaker.AI:
                prompt += "<|assistant|>\n"
                prompt += turn.text + "\n"

        # Append the next prompt
        last_prompt_is_user = True
        if conversation.turns[-1].speaker in (Speaker.USER, Speaker.SYSTEM):
            prompt += "<|assistant|>\n"
            last_prompt_is_user = True
        else:
            logging.warning(
                "The last turn in the conversation is from the AI, however this model does not (usually) support continuing conversation from the user perspective."
            )
            prompt += "<|user|>\n"
            last_prompt_is_user = False

        return prompt, last_prompt_is_user

    def _extract_last_turn(self, outputs: List[str], last_prompt_is_user: bool) -> List[str]:
        return [o.split("<|assistant|>" if last_prompt_is_user else "<|user|>")[0].strip() for o in outputs]


@register_engine
@singleton
class Koala7B(KoalaLMEngine):
    name = ("Koala (7B)", "koala-7B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("koala_7B", "KOALA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("KOALA_WEIGHTS_ROOT", None), "koala_7B")


@register_engine
@singleton
class Koala13BV1(KoalaLMEngine):
    name = ("Koala (13B, v1)", "koala-13B-v1")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("koala_13B_v1", "KOALA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("KOALA_WEIGHTS_ROOT", None), "koala_13B_v1")


@register_engine
@singleton
class Koala13BV2(KoalaLMEngine):
    name = ("Koala (13B, v2)", "koala-13B-v2")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("koala_13B_v2", "KOALA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("KOALA_WEIGHTS_ROOT", None), "koala_13B_v2")


@register_engine
@singleton
class Vicuna7B(Vicuna11LMEngine):
    name = ("Vicuna (7B)", "vicuna-7B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("vicuna_7B", "VICUNA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("VICUNA_WEIGHTS_ROOT", None), "vicuna_7B")


@register_engine
@singleton
class Vicuna13B(Vicuna11LMEngine):
    name = ("Vicuna (13B)", "vicuna-13B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("vicuna_13B", "VICUNA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("VICUNA_WEIGHTS_ROOT", None), "vicuna_13B")


@register_engine
@singleton
class Alpaca13B(AlpacaLMEngine):
    name = ("Alpaca (13B)", "alpaca-13B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("alpaca_13B", "ALPACA_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALPACA_WEIGHTS_ROOT", None), "alpaca_13B")


@register_engine
@singleton
class Tulu7B(AllenAILlamaLMEngine):
    name = ("Tulu (7B)", "tulu-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tulu-7b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "tulu-7b")


@register_engine
@singleton
class Tulu13B(AllenAILlamaLMEngine):
    name = ("Tulu (13B)", "tulu-13b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tulu-13b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "tulu-13b")


@register_engine
@singleton
class Tulu30B(AllenAILlamaLMEngine):
    name = ("Tulu (30B)", "tulu-30b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tulu-30b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "tulu-30b")


@register_engine
@singleton
class Tulu65B(AllenAILlamaLMEngine):
    name = ("Tulu (65B)", "tulu-65b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tulu-65b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "tulu-65b")


@register_engine
@singleton
class OIShareGPT7B(AllenAILlamaLMEngine):
    name = ("Open Instruct ShareGPT (7B)", "open-instruct-sharegpt-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("open-instruct-sharegpt-7b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "open-instruct-sharegpt-7b")


@register_engine
@singleton
class OIShareGPT13B(AllenAILlamaLMEngine):
    name = ("Open Instruct ShareGPT (13B)", "open-instruct-sharegpt-13b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("open-instruct-sharegpt-13b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "open-instruct-sharegpt-13b")


@register_engine
@singleton
class OIShareGPT30B(AllenAILlamaLMEngine):
    name = ("Open Instruct ShareGPT (30B)", "open-instruct-sharegpt-30b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("open-instruct-sharegpt-30b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "open-instruct-sharegpt-30b")


@register_engine
@singleton
class OIShareGPT65B(AllenAILlamaLMEngine):
    name = ("Open Instruct ShareGPT (65B)", "open-instruct-sharegpt-65b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("open-instruct-sharegpt-65b", "ALLENAI_WEIGHTS_ROOT", device=device)

    @staticmethod
    def is_configured() -> bool:
        return LlamaLMEngine._is_configured(os.environ.get("ALLENAI_WEIGHTS_ROOT", None), "open-instruct-sharegpt-65b")



import copy
import logging
from typing import Any, List, Optional, Type

import torch
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    TFAutoModelForCausalLM,
    is_tf_available,
    is_torch_available,
)
from transformers.pipelines import PIPELINE_REGISTRY, TextGenerationPipeline, pipeline

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.huggingface import check_huggingface_model_files_are_local
from grazier.utils.python import singleton

# Some models are too new, and not directly supported by the text-generation pipeline, but do have support for
# AutoModelForCausalLM. We can register a new pipeline that doesn't check the model type, and use that instead.
_tgp = copy.copy(TextGenerationPipeline)
_tgp.check_model_type = lambda *args, **kwargs: None
PIPELINE_REGISTRY.register_pipeline(
    task="text-generation-no-check-models",
    pipeline_class=_tgp,
    tf_model=(TFAutoModelForCausalLM,) if is_tf_available() else (),
    pt_model=(AutoModelForCausalLM,) if is_torch_available() else (),
    default={"model": {"pt": ("t5-base", "686f1db"), "tf": ("t5-base", "686f1db")}},
    type="text",
)


class HuggingFaceTextGenerationLMEngine(LLMEngine):
    def __init__(
        self,
        model: str,
        device: Optional[str] = None,
        override_check_models: bool = False,
        quantize: bool = False,
    ):
        super().__init__(device=device)

        # Setup quantization
        if quantize:
            nf4_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_use_double_quant=True,
                bnb_4bit_compute_dtype=torch.float16,
            )
        else:
            nf4_config = None

        self._model = AutoModelForCausalLM.from_pretrained(
            model,
            device_map="auto" if not self.device_specified else None,
            quantization_config=nf4_config,
            trust_remote_code=True,
        )
        self._tokenizer = AutoTokenizer.from_pretrained(
            model,
            trust_remote_code=True,
            padding_side="left",
        )
        self._tokenizer.pad_token_id = self._model.config.eos_token_id or self._tokenizer.eos_token_id

        self._generator = pipeline(
            "text-generation" if not override_check_models else "text-generation-no-check-models",
            model=self._model,
            tokenizer=self._tokenizer,
            framework="pt",
            device=self.device if self.device_specified else None,
            device_map="auto" if not self.device_specified else None,
            torch_dtype=torch.float16 if quantize else None,
            trust_remote_code=True,
        )

    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        # Handle the kwargs
        _params = {
            "max_new_tokens": kwargs.get("max_new_tokens", kwargs.pop("max_tokens", 256)),
            "min_new_tokens": kwargs.get("min_new_tokens", 10),
            "num_return_sequences": n_completions,
            "temperature": kwargs.get("temperature", None),
            "return_full_text": False,
        } | kwargs

        outputs = self._generator(
            prompt,
            do_sample=n_completions > 1,
            pad_token_id=self._tokenizer.pad_token_id,
            eos_token_id=self._tokenizer.eos_token_id,
            **_params,
        )

        outputs = [g["generated_text"].strip() for g in outputs]  # type: ignore

        return outputs

    @staticmethod
    def from_hub_model(modelstr: str) -> Type["HuggingFaceTextGenerationLMEngine"]:
        class _RemoteHFModel(HuggingFaceTextGenerationLMEngine):
            name = (modelstr, modelstr)

            def __init__(self, device: Optional[str] = None):
                super().__init__(modelstr, device=device)

        _cf_class = singleton(_RemoteHFModel)
        return _cf_class  # type: ignore

    @staticmethod
    def requires_configuration() -> bool:
        # Check if there's an internet connection, if so, then we need to check if the model is available
        # locally. If not, then we can't use this engine.
        import requests

        try:
            requests.get("https://huggingface.co/")
            return False
        except requests.exceptions.ConnectionError:
            return True
        except Exception as e:
            logging.warning(f"Unexpected error when checking if HuggingFace is available: {e}")
            return True


@register_engine
@singleton
class GPTJ6B(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-J 6B", "gptj-6B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("EleutherAI/gpt-j-6B", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("EleutherAI/gpt-j-6B")


@register_engine
@singleton
class GPT2(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-2", "gpt2")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("gpt2", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("gpt2")


@register_engine
@singleton
class GPT2Med(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-2 Medium", "gpt2-med")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("gpt2-medium", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("gpt2-medium")


@register_engine
@singleton
class GPT2Lg(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-2 Large", "gpt2-lg")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("gpt2-large", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("gpt2-large")


@register_engine
@singleton
class GPT2XL(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-2 XL", "gpt2-xl")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("gpt2-xl", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("gpt2-xl")


@register_engine
@singleton
class DistilGPT2(HuggingFaceTextGenerationLMEngine):
    name = ("DistilGPT-2", "distilgpt2")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("distilgpt2", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("distilgpt2")


@register_engine
@singleton
class GPTNeo125M(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-Neo 125M", "gptneo-125M")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("EleutherAI/gpt-neo-125M", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("EleutherAI/gpt-neo-125M")


@register_engine
@singleton
class GPTNeo1B(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-Neo 1.3B", "gptneo-1.3B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("EleutherAI/gpt-neo-1.3B", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("EleutherAI/gpt-neo-1.3B")


@register_engine
@singleton
class GPTNeo2B(HuggingFaceTextGenerationLMEngine):
    name = ("GPT-Neo 2.7B", "gptneo-2.7B")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("EleutherAI/gpt-neo-2.7B", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("EleutherAI/gpt-neo-2.7B")


@register_engine
@singleton
class StableLMBase3B(HuggingFaceTextGenerationLMEngine):
    name = ("StableLM 3B", "stablelm-3b")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("stabilityai/stablelm-base-alpha-3b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("stabilityai/stablelm-base-alpha-3b")


@register_engine
@singleton
class StableLMBase7B(HuggingFaceTextGenerationLMEngine):
    name = ("StableLM 7B", "stablelm-7b")

    def __init__(self, device: Optional[str] = "defer") -> None:
        super().__init__("stabilityai/stablelm-base-alpha-7b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("stabilityai/stablelm-base-alpha-7b")


@register_engine
@singleton
class OPT125M(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 125M", "opt-125M")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-125m", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-125m")


@register_engine
@singleton
class OPT350M(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 350M", "opt-350M")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-350m", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-350m")


@register_engine
@singleton
class OPT1x3B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 1.3B", "opt-1.3b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-1.3b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-1.3b")


@register_engine
@singleton
class OPT2x7B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 2.7B", "opt-2.7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-2.7b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-2.7b")


@register_engine
@singleton
class OPT6x7B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 6.7B", "opt-6.7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-6.7b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-6.7b")


@register_engine
@singleton
class OPT13B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 13B", "opt-13b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-13b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-13b")


@register_engine
@singleton
class OPT30B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 30B", "opt-30b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-30b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-30b")


@register_engine
@singleton
class OPT66B(HuggingFaceTextGenerationLMEngine):
    name = ("OPT 66B", "opt-66b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("facebook/opt-66b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("facebook/opt-66b")


@register_engine
@singleton
class Falcon40B(HuggingFaceTextGenerationLMEngine):
    name = ("Falcon 40B", "falcon-40b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tiiuae/falcon-40b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("tiiuae/falcon-40b")


@register_engine
@singleton
class Falcon7B(HuggingFaceTextGenerationLMEngine):
    name = ("Falcon 7B", "falcon-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tiiuae/falcon-7b", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("tiiuae/falcon-7b")


@register_engine
@singleton
class FalconRW7B(HuggingFaceTextGenerationLMEngine):
    name = ("Falcon RW 7B", "falcon-rw-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tiiuae/falcon-rw-7b", device=device, override_check_models=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("tiiuae/falcon-rw-7b")


@register_engine
@singleton
class FalconRW1B(HuggingFaceTextGenerationLMEngine):
    name = ("Falcon RW 1B", "falcon-rw-1b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("tiiuae/falcon-rw-1b", device=device, override_check_models=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("tiiuae/falcon-rw-1b")


@register_engine
@singleton
class H2OGPT_GM_OASST1_EN_2048_FALCON7B(HuggingFaceTextGenerationLMEngine):  # noqa: N801
    name = ("H2OGPT_GM_OASST1_EN_2048_FALCON7B", "h2ogpt-gm-oasst1-en-2048-falcon-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("h2oai/h2ogpt-gm-oasst1-en-2048-falcon-7b", device=device, override_check_models=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("h2oai/h2ogpt-gm-oasst1-en-2048-falcon-7b")

    @property
    def prompt_prefix(self) -> str:
        return "<|prompt|>"

    @property
    def prompt_suffix(self) -> str:
        return "<|endoftext|><|answer|>"


@register_engine
@singleton
class LLama27B(HuggingFaceTextGenerationLMEngine):
    name = ("Llama-2 (7B)", "llama-2-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("meta-llama/Llama-2-7b-hf", device=device, quantize=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("meta-llama/Llama-2-7b-hf")


@register_engine
@singleton
class LLama213B(HuggingFaceTextGenerationLMEngine):
    name = ("Llama-2 (13B)", "llama-2-13b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("meta-llama/Llama-2-13b-hf", device=device, quantize=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("meta-llama/Llama-2-13b-hf")


@register_engine
@singleton
class LLama270B(HuggingFaceTextGenerationLMEngine):
    name = ("Llama-2 (70B)", "llama-2-70b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("meta-llama/Llama-2-70b-hf", device=device, quantize=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("meta-llama/Llama-2-70b-hf")


@register_engine
@singleton
class Mistral7B(HuggingFaceTextGenerationLMEngine):
    name = ("Mistral (7B)", "mistral-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("mistralai/Mistral-7B-v0.1", device=device, quantize=True)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("mistralai/Mistral-7B-v0.1")



import logging

import pytest
from PIL import Image

import grazier


@pytest.mark.parametrize(
    "engine",
    [
        "OpenFlamingo-3B-vitl-mpt1b",
    ],
)
def test_open_flamingo_ilm_engine(engine: str) -> None:
    _engine = grazier.get(engine, type="image")

    # Load the test image
    try:
        image = Image.open("test_data/dog.jpg")
    except FileNotFoundError:
        pytest.skip("Test image not found")

    # Generate a caption
    caption = _engine(image)

    # Check that the caption is not empty
    assert len(caption[0].strip()) > 0, f"Caption is empty: {caption}"

    # Check that the caption contains the word "llama"
    if "dog" not in caption[0].lower() and "puppy" not in caption[0].lower():
        logging.warning(f'Word "dog/puppy" not found in caption "{caption}"')

    print(caption)


if __name__ == "__main__":
    test_open_flamingo_ilm_engine("OpenFlamingo-9B-vitl-mpt7b")



import logging
import os
from typing import Any, List

import anthropic

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import retry, singleton


class AnthropicLMEngine(LLMChat):
    def __init__(self, model: str):
        super().__init__(device="api")
        self._client = anthropic.Client(api_key=os.getenv("ANTHROPIC_API_KEY", None))
        self._model = model

    @retry()
    def _completion(self, prompt: str, **kwargs: Any) -> Any:
        kwargs = {
            "temperature": kwargs.get("temperature", 0.7),
            "max_tokens_to_sample": kwargs.get("max_tokens_to_sample", kwargs.pop("max_tokens", 256)),
            "model": self._model,
            "prompt": prompt,
        } | kwargs

        return self._client.completion(
            **kwargs,
        )

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        # Some odd anthropic assertions
        if conversation.turns[-1].speaker != Speaker.USER:
            raise AssertionError("Last turn must be a user turn")
        # Assert that conversations altrenate between user and AI (anthropic doesn't support system turns)
        for idx, turn in enumerate([c for c in conversation.turns if c.speaker != Speaker.SYSTEM]):
            if idx % 2 == 0 and turn.speaker != Speaker.USER:
                raise AssertionError("Conversations must alternate between user and AI turns")
            if idx % 2 == 1 and turn.speaker != Speaker.AI:
                raise AssertionError("Conversations must alternate between user and AI turns")

        # Construct the messages list from the conversation
        prompt = ""
        for idx, turn in enumerate(conversation.turns):
            if turn.speaker == Speaker.SYSTEM:
                logging.warning("Anthropic does not support SYSTEM turns, skipping...")
            elif turn.speaker == Speaker.USER:
                prompt += f"{anthropic.HUMAN_PROMPT} "
                prompt += turn.text + " "
            elif turn.speaker == Speaker.AI:
                prompt += f"{anthropic.AI_PROMPT} "
                prompt += turn.text + " "

        # add the last turn
        prompt += f"{anthropic.AI_PROMPT}"

        temperature = kwargs.get("temperature", 0.7)

        samples = []
        for _ in range(n_completions):
            resp = self._completion(prompt, temperature=temperature)
            samples.append(resp["completion"])

        return [ConversationTurn(text=s.strip(), speaker=Speaker.AI) for s in samples]

    @staticmethod
    def is_configured() -> bool:
        return os.getenv("ANTHROPIC_API_KEY", None) is not None


@register_engine
@singleton
class Claude(AnthropicLMEngine):
    name = ("Claude", "claude")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("claude-1")


@register_engine
@singleton
class Claude100K(AnthropicLMEngine):
    name = ("Claude 100K", "claude-100k")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("claude-1-100k")


@register_engine
@singleton
class ClaudeInstant(AnthropicLMEngine):
    name = ("Claude Instant", "claude-instant")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("claude-instant-1")


@register_engine
@singleton
class ClaudeInstant100K(AnthropicLMEngine):
    name = ("Claude Instant 100K", "claude-instant-100k")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("claude-instant-1-100k")


@register_engine
@singleton
class Claude2(AnthropicLMEngine):
    name = ("Claude 2", "claude-2")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("claude-2")



import functools
import hashlib
import logging
import os
import time
from contextlib import AbstractContextManager
from typing import Any, Callable, Collection, Generic, List, Optional, Type, TypeVar


def compute_md5_hash_from_bytes(input_bytes: bytes) -> str:
    """Compute the MD5 hash of a byte array."""
    return str(hashlib.md5(input_bytes).hexdigest())


class chdir(AbstractContextManager):  # noqa: N801
    """Non thread-safe context manager to change the current working directory."""

    def __init__(self, path: str) -> None:
        self.path = path
        self._old_cwd: List[str] = []

    def __enter__(self) -> None:
        self._old_cwd.append(os.getcwd())
        os.chdir(self.path)

    def __exit__(self, *excinfo) -> None:  # type: ignore
        os.chdir(self._old_cwd.pop())


Q = TypeVar("Q", bound=Callable[..., Any])


def retry(no_retry_on: Optional[Collection[Type[Exception]]] = None) -> Callable[[Q], Q]:
    """Retry a function if it throws an exception."""

    def decorator(func: Q) -> Q:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            error = None
            for backoff in [0.1, 0.2, 0.5, 1.0, 2.0, 5.0, 10.0]:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if no_retry_on is not None and type(e) in no_retry_on:
                        raise e

                    # Backoff and try again
                    time.sleep(backoff)
                    logging.warn(f"Retrying {func.__name__} after {backoff}s due to {e}")
                    error = e
                    continue

            if error is not None:
                raise error  # type: ignore

        return wrapper  # type: ignore

    return decorator


T = TypeVar("T")
S = TypeVar("S")


class _SingletonWrapper(Generic[T]):
    def __init__(self, cls: Type[T]):
        self.__wrapped__ = cls
        self._instance: Optional[T] = None
        functools.update_wrapper(self, cls)

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Returns a single instance of decorated class"""
        if self._instance is None:
            self._instance = self.__wrapped__(*args, **kwargs)
        return self._instance

    # Pass class methods/attributes through to the wrapped class
    def __getattr__(self, attr: str) -> Any:
        if hasattr(self.__wrapped__, attr):
            return getattr(self.__wrapped__, attr)
        raise AttributeError(f"{self.__wrapped__.__name__} has no attribute {attr}")


def singleton(cls: Type[S]) -> _SingletonWrapper[S]:
    return _SingletonWrapper(cls)



import logging
import random

import pytest
import torch

from grazier.engines.llm import LLMEngine


@pytest.mark.parametrize(
    "engine",
    ["gpt2", "stablelm-3B", "falcon-7b", "llama-2-7b", "mistral-7b"],
)
def test_huggingface_llm_engine(engine: str) -> None:
    try:
        _engine = LLMEngine.from_string(engine)
        random_number = random.randint(0, 100)
        response = _engine(f"My name, followed by a colon with the number {random_number} is:")

        for r in response:
            if len(r.strip()) == 0:
                pytest.skip(f"Empty response from {engine} engine")
            if str(random_number) not in r:
                logging.warning(f'Number "{random_number}" not found in response "{r}"')
    except torch.cuda.OutOfMemoryError:
        pytest.skip(f"Out of memory error from {engine} engine")



import logging

import click
from rich import print
from rich.logging import RichHandler


@click.group()
def main() -> None:
    # Configure logging
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )


@main.command()
@click.option("--configured", is_flag=True, help="Only list configured engines")
def list_engines(
    configured: bool,
) -> None:
    """List available engines."""
    from grazier.engines.chat import LM_CHAT_ENGINES_CLI
    from grazier.engines.llm import LM_ENGINES_CLI

    # Types here are super broken because of the wrapped decorator
    configured_models_llm, unconfigured_models_llm = [], []
    for k, v in sorted(LM_ENGINES_CLI.items()):  # type: ignore
        engine_requires_configuration = v.requires_configuration()
        engine_is_configured = not engine_requires_configuration or v.is_configured()
        if configured and not engine_is_configured:
            continue
        configured_models_llm.append((k, v)) if engine_is_configured else unconfigured_models_llm.append((k, v))

    # Types here are super broken because of the wrapped decorator
    configured_models_chat, unconfigured_models_chat = [], []
    for k, v in sorted(LM_CHAT_ENGINES_CLI.items()):  # type: ignore
        engine_requires_configuration = v.requires_configuration()
        engine_is_configured = not engine_requires_configuration or v.is_configured()
        if configured and not engine_is_configured:
            continue
        configured_models_chat.append((k, v)) if engine_is_configured else unconfigured_models_chat.append((k, v))

    print("Available Models (Completions API, Configured):")
    print("\n".join([f"\t [green]- {v.name[0]}/{k} [/green]" for k, v in configured_models_llm]))  # type: ignore
    if not configured:
        print("\n".join([f"\t [red]- {v.name[0]}/{k} (Unconfigured) [/red]" for k, v in unconfigured_models_llm]))  # type: ignore

    print()
    print("Available Models (Chat API):")
    print("\n".join([f"\t [green]- {v.name[0]}/{k} [/green]" for k, v in configured_models_chat]))  # type: ignore
    if not configured:
        print("\n".join([f"\t [red]- {v.name[0]}/{k} (Unconfigured) [/red]" for k, v in unconfigured_models_chat]))  # type: ignore


@main.command()
@click.argument("engine")
@click.argument("prompt")
def complete(
    engine: str,
    prompt: str,
) -> None:
    from grazier import LLMEngine

    _e = LLMEngine.from_string(engine)
    print(_e(prompt, n_completions=2)[0])


@main.command()
@click.argument("engine")
def configure(
    engine: str,
) -> None:
    """List available engines."""
    from grazier.engines.chat import LM_CHAT_ENGINES_CLI
    from grazier.engines.llm import LM_ENGINES_CLI

    if engine.lower() in LM_ENGINES_CLI:
        LM_ENGINES_CLI[engine.lower()].configure()
    elif engine.lower() in LM_CHAT_ENGINES_CLI:
        LM_CHAT_ENGINES_CLI[engine.lower()].configure()
    else:
        logging.error(f"Engine {engine} not found.")


@main.command()
def reset_credential_store():
    from grazier.utils.secrets import reset_credential_store

    reset_credential_store()



import logging
from typing import Any, List, Optional

import torch
from huggingface_hub import hf_hub_download
from PIL import Image

from grazier.engines.image import ILMEngine, register_engine
from grazier.utils.python import singleton

try:
    from open_flamingo import create_model_and_transforms
except ImportError:
    create_model_and_transforms = None


class OpenFlamingoILMEngine(ILMEngine):
    def __init__(
        self,
        checkpoint_str: str,
        vision_encoder: str,
        language_model: str,
        cross_attn_every_n_layers=1,
        device: Optional[str] = None,
        quantize: bool = False,
    ) -> None:
        super().__init__(device=device)

        if create_model_and_transforms is None:
            raise ImportError("OpenFlamingo is not installed. Please install it with `pip install open-flamingo`.")

        if quantize:
            logging.warning("Quantization is not supported for OpenFlamingo models.")

        self._model, self._image_processor, self._tokenizer = create_model_and_transforms(
            clip_vision_encoder_path=vision_encoder,
            clip_vision_encoder_pretrained="openai",
            lang_encoder_path=language_model,
            tokenizer_path=language_model,
            cross_attn_every_n_layers=cross_attn_every_n_layers,
        )
        self._model = self._model.to(self.device)

        checkpoint_path = hf_hub_download(checkpoint_str, "checkpoint.pt")
        self._model.load_state_dict(torch.load(checkpoint_path), strict=False)

        self._tokenizer.paddding_side = "left"

    def call(self, image: Image, prompt: Optional[str] = None, n_completions: int = 1, **kwargs: Any) -> List[str]:
        if prompt is None:
            prompt = "<image>A photo of"
        if "<image>" not in prompt:
            prompt = "<image>" + prompt
        if prompt.count("<image>") > 1:
            raise ValueError("Prompt must contain at most one <image> token.")

        processed_image = (
            torch.cat(
                [
                    self._image_processor(image).unsqueeze(0),
                ],
                dim=0,
            )
            .unsqueeze(1)
            .unsqueeze(0)
        )

        text_inputs = self._tokenizer([prompt], return_tensors="pt")

        # Move the inputs to the correct device
        processed_image = processed_image.to(self.device)
        text_inputs = {k: v.to(self.device) for k, v in text_inputs.items() if isinstance(v, torch.Tensor)}

        outputs = self._model.generate(
            vision_x=processed_image,
            lang_x=text_inputs["input_ids"],
            attention_mask=text_inputs["attention_mask"],
            max_new_tokens=kwargs.get("max_new_tokens", kwargs.pop("max_tokens", 128)),
            num_return_sequences=n_completions,
            do_sample=n_completions > 1,
            pad_token_id=self._tokenizer.pad_token_id,
            eos_token_id=self._tokenizer.eos_token_id,
            **kwargs,
        )

        # Remove the prompt from the output
        outputs = outputs[:, text_inputs["input_ids"].shape[-1] :]
        outputs = self._tokenizer.batch_decode(outputs, skip_special_tokens=True)

        return [o.strip() for o in outputs]


@register_engine
@singleton
class OpenFlamingo3B(OpenFlamingoILMEngine):
    name = ("Open-Flamingo (3B, ViT-L, MPT-1B)", "OpenFlamingo-3B-vitl-mpt1b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(
            checkpoint_str="openflamingo/OpenFlamingo-3B-vitl-mpt1b",
            vision_encoder="ViT-L-14",
            language_model="mosaicml/mpt-1b-redpajama-200b",
            device=device,
            quantize=quantize,
        )

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class OpenFlamingo3BInstruct(OpenFlamingoILMEngine):
    name = ("Open-Flamingo (3B, ViT-L, MPT-1B-Dolly)", "OpenFlamingo-3B-vitl-mpt1b-dolly")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(
            checkpoint_str="openflamingo/OpenFlamingo-3B-vitl-mpt1b-langinstruct",
            vision_encoder="ViT-L-14",
            language_model="mosaicml/mpt-1b-redpajama-200b-dolly",
            device=device,
            quantize=quantize,
        )

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class OpenFlamingo9B(OpenFlamingoILMEngine):
    name = ("Open-Flamingo (9B, ViT-L, MPT-7B)", "OpenFlamingo-9B-vitl-mpt7b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(
            checkpoint_str="openflamingo/OpenFlamingo-9B-vitl-mpt7b",
            vision_encoder="ViT-L-14",
            language_model="anas-awadalla/mpt-7b",
            cross_attn_every_n_layers=4,
            device=device,
            quantize=quantize,
        )

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class OpenFlamingo4B(OpenFlamingoILMEngine):
    name = ("Open-Flamingo (4B, ViT-L, INCITE-Base-3B)", "OpenFlamingo-4B-vitl-rpj3b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(
            checkpoint_str="openflamingo/OpenFlamingo-4B-vitl-rpj3b",
            vision_encoder="ViT-L-14",
            language_model="atogethercomputer/RedPajama-INCITE-Base-3B-v1",
            cross_attn_every_n_layers=2,
            device=device,
            quantize=quantize,
        )

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class OpenFlamingo4BInstruct(OpenFlamingoILMEngine):
    name = ("Open-Flamingo (4B, ViT-L, INCITE-Instruct-3B)", "OpenFlamingo-4B-vitl-rpj3b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(
            checkpoint_str="openflamingo/OpenFlamingo-4B-vitl-rpj3b-langinstruct",
            vision_encoder="ViT-L-14",
            language_model="togethercomputer/RedPajama-INCITE-Instruct-3B-v1",
            cross_attn_every_n_layers=2,
            device=device,
            quantize=quantize,
        )

    @staticmethod
    def is_configured() -> bool:
        return True



#    Copyright 2023 Rohan Taori, Ishaan Gulrajani, Tianyi Zhang, Yann Dubois, Xuechen Li
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Dict, Optional

import fire
import torch
import tqdm
import transformers


def smart_tokenizer_and_embedding_resize(
    special_tokens_dict: Dict,
    tokenizer: transformers.PreTrainedTokenizer,
    model: transformers.PreTrainedModel,
):
    """Resize tokenizer and embedding.

    Note: This is the unoptimized version that may make your embedding size not be divisible by 64.
    """
    num_new_tokens = tokenizer.add_special_tokens(special_tokens_dict)
    model.resize_token_embeddings(len(tokenizer))

    if num_new_tokens > 0:
        input_embeddings = model.get_input_embeddings().weight.data
        output_embeddings = model.get_output_embeddings().weight.data

        input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)
        output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)

        input_embeddings[-num_new_tokens:] = input_embeddings_avg
        output_embeddings[-num_new_tokens:] = output_embeddings_avg


@torch.inference_mode()
def make_diff(
    path_raw: str,
    path_tuned: str,
    path_diff: str,
    device="cpu",  # "cuda" or "cpu"
):
    """Make the weight diff.

    This function is given to present full transparency of how the weight diff was created.

    Run:
        python weight_diff.py make_diff --path_raw <your_path_raw> --path_tuned <your_path_tuned> --path_diff <your_path_diff>
    """
    model_tuned: transformers.PreTrainedModel = transformers.AutoModelForCausalLM.from_pretrained(
        path_tuned,
        device_map={"": torch.device(device)},
        torch_dtype=torch.float32,
        low_cpu_mem_usage=True,
    )
    model_raw: transformers.PreTrainedModel = transformers.AutoModelForCausalLM.from_pretrained(
        path_raw,
        device_map={"": torch.device(device)},
        torch_dtype=torch.float32,
        low_cpu_mem_usage=True,
    )

    tokenizer_tuned: transformers.PreTrainedTokenizer = transformers.AutoTokenizer.from_pretrained(path_tuned)
    tokenizer_raw: transformers.PreTrainedTokenizer = transformers.AutoTokenizer.from_pretrained(path_raw)
    if tokenizer_raw.pad_token is None:
        tokenizer_raw.add_special_tokens(dict(pad_token="[PAD]"))
        model_raw.resize_token_embeddings(len(tokenizer_raw))

    state_dict_tuned = model_tuned.state_dict()
    state_dict_raw = model_raw.state_dict()
    for key in tqdm.tqdm(state_dict_tuned):
        state_dict_tuned[key].add_(-state_dict_raw[key])

    model_tuned.save_pretrained(path_diff)
    tokenizer_tuned.save_pretrained(path_diff)


@torch.inference_mode()
def recover(
    path_raw,
    path_diff,
    path_tuned: Optional[str] = None,
    original_model: Optional[str] = None,
    device="cpu",
    test_inference=True,
):
    """Recover the original weights from the released weight diff.

    This function is given for you to run.

    Things to do before running this:
        1. Convert Meta's released weights into huggingface format. Follow this guide:
            https://huggingface.co/docs/transformers/main/model_doc/llama
        2. Make sure you cloned the released weight diff into your local machine. The weight diff is located at:
            https://huggingface.co/tatsu-lab/alpaca-7b/tree/main
        3. Run this function with the correct paths. E.g.,
            python weight_diff.py recover --path_raw <path_to_step_1_dir> --path_diff <path_to_step_2_dir>

    Additional notes:
        - If things run too slowly, and you have an 80G GPU lying around, let GPU go brrr by setting `--device "cuda"`.
        - If you want to save the recovered weights, set `--path_tuned <your_path_tuned>`.
            Next time you can load the recovered weights directly from `<your_path_tuned>`.
        - to run inference on a reference model (e.g. to ensure diff is correct), set `--original_model <your_model_name>`.
    """
    model_raw: transformers.PreTrainedModel = transformers.AutoModelForCausalLM.from_pretrained(
        path_raw,
        device_map={"": torch.device(device)},
        torch_dtype=torch.float32,
        low_cpu_mem_usage=True,
    )
    model_recovered: transformers.PreTrainedModel = transformers.AutoModelForCausalLM.from_pretrained(
        path_diff,
        device_map={"": torch.device(device)},
        torch_dtype=torch.float32,
        low_cpu_mem_usage=True,
    )

    tokenizer_raw: transformers.PreTrainedTokenizer = transformers.LlamaTokenizer.from_pretrained(path_raw)
    if tokenizer_raw.pad_token is None:
        smart_tokenizer_and_embedding_resize(
            special_tokens_dict=dict(pad_token="[PAD]"),
            model=model_raw,
            tokenizer=tokenizer_raw,
        )
    tokenizer_recovered: transformers.PreTrainedTokenizer = transformers.LlamaTokenizer.from_pretrained(path_diff)

    state_dict_recovered = model_recovered.state_dict()
    state_dict_raw = model_raw.state_dict()
    for key in tqdm.tqdm(state_dict_recovered):
        state_dict_recovered[key].add_(state_dict_raw[key])

    if path_tuned is not None:
        model_recovered.save_pretrained(path_tuned)
        tokenizer_recovered.save_pretrained(path_tuned)

    if test_inference:
        input_text = (
            "Below is an instruction that describes a task. "
            "Write a response that appropriately completes the request.\r\n\r\n"
            "### Instruction:\r\nList three technologies that make life easier.\r\n\r\n### Response:"
        )
        inputs = tokenizer_recovered(input_text, return_tensors="pt")
        out = model_recovered.generate(inputs=inputs.input_ids, max_new_tokens=100)
        output_text = tokenizer_recovered.batch_decode(out, skip_special_tokens=True)[0]
        output_text = output_text[len(input_text) :]
        print("Recovered model:")
        print(f"Input: {input_text}\nCompletion: {output_text}")
        if original_model:
            og_tokenizer = transformers.AutoTokenizer.from_pretrained(original_model)
            og_model = transformers.AutoModelForCausalLM.from_pretrained(original_model)
            og_inputs = og_tokenizer(input_text, return_tensors="pt")
            og_out = og_model.generate(inputs=og_inputs.input_ids, max_new_tokens=100)
            og_output_text = og_tokenizer.batch_decode(og_out, skip_special_tokens=True)[0]
            og_output_text = og_output_text[len(input_text) :]
            print("Original model:")
            print(f"Input: {input_text}\nCompletion: {og_output_text}")

    return model_recovered, tokenizer_recovered


def main(task, **kwargs):
    globals()[task](**kwargs)


if __name__ == "__main__":
    fire.Fire(main)



import logging
from abc import abstractmethod
from typing import Any, List

import openai
from rich.prompt import Prompt

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.python import retry, singleton
from grazier.utils.secrets import get_secret, set_secret


def _setup_api_keys():
    # Setup openai api keys
    openai.api_key = get_secret("OPENAI_API_KEY", None)
    openai.organization = get_secret("OPENAI_API_ORG", None)


@singleton
class OpenAI:
    USAGE = 0.0


class OpenAICompletionLLMEngine(LLMEngine):
    @property
    @abstractmethod
    def cost_per_token(self) -> float:
        raise NotImplementedError()

    def __init__(self, model: str):
        self._model = model

        _setup_api_keys()

        super().__init__(device="api")

    @retry(no_retry_on=(openai.error.AuthenticationError,))
    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        cp = openai.Completion.create(model=self._model, prompt=prompt, n=n_completions, **kwargs)  # type: ignore
        OpenAI.USAGE += int(cp.usage.total_tokens) * self.cost_per_token
        return [i.text for i in cp.choices]  # type: ignore

    @staticmethod
    def is_configured() -> bool:
        _setup_api_keys()
        return openai.api_key is not None

    @staticmethod
    def configure():
        if OpenAICompletionLLMEngine.is_configured():
            reconfigure = Prompt.ask("OpenAI API key is already configured. Reconfigure?", choices=["y", "n"])
            if reconfigure == "n":
                logging.info("OpenAI API key already configured.")
                return

        have_openai_key = Prompt.ask("Do you have an OpenAI API key?", choices=["y", "n"])
        if have_openai_key == "n":
            Prompt.ask(
                "Please follow the instructions here [u cyan]https://www.howtogeek.com/885918/how-to-get-an-openai-api-key/[/u cyan] to obtain an API key.\nWhen you have the key, press enter to continue",
            )

        openai_key = Prompt.ask("Please enter your OpenAI API key (It will be saved to your platform's secrets store)")
        set_secret("OPENAI_API_KEY", openai_key.strip())

        # Save the openai api key to the secrets store
        logging.info("OpenAI API successfully configured.")


@register_engine
@singleton
class GPT3Davinci3(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Davinci", "gpt3-davinci3")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("text-davinci-003")


@register_engine
@singleton
class GPT3Davinci2(OpenAICompletionLLMEngine):
    cost_per_token = 0.03 / 1000
    name = ("GPT-3 Davinci", "gpt3-davinci2")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("text-davinci-002")


@register_engine
@singleton
class GPT3Curie(OpenAICompletionLLMEngine):
    cost_per_token = 0.002 / 1000
    name = ("GPT-3 Curie", "gpt3-curie")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("text-curie-001")


@register_engine
@singleton
class GPT3Babbage(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Babbage", "gpt3-babbage")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("text-babbage-001")


@register_engine
@singleton
class GPT3Ada(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Ada", "gpt3-ada")

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("text-ada-001")



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Type, TypeVar

from PIL import Image

from grazier.engines.default import Engine
from grazier.utils.pytorch import select_device


class ILMEngine(Engine):
    def __init__(self, device: Optional[str] = None) -> None:
        self.device = select_device(device)
        self.device_specified = device is not None

    @property
    def prompt_prefix(self) -> str:
        return ""

    @property
    def prompt_suffix(self) -> str:
        return ""

    def __call__(self, image: Image, prompt: Optional[str] = None, n_completions: int = 1, **kwargs: Any) -> List[str]:
        if prompt is not None:
            prompt = self.prompt_prefix + prompt + self.prompt_suffix
        return self.call(image, prompt, n_completions, **kwargs)

    @abstractmethod
    def call(self, image: Image, prompt: Optional[str] = None, n_completions: int = 1, **kwargs: Any) -> List[str]:
        raise NotImplementedError()

    def __repr__(
        self,
    ) -> str:
        return f"{self.__class__.__name__}({self.name[0]})"

    @staticmethod
    def from_string(typestr: str, **kwargs: Any) -> "ILMEngine":
        typestr = typestr.lower()

        if typestr in ILM_ENGINES:
            if not ILM_ENGINES[typestr].is_configured() and ILM_ENGINES[typestr].requires_configuration():
                raise ValueError(
                    f"Image/Language model type: {typestr} requires configuration. Please run `grazier configure {typestr}` first."
                )
            else:
                return ILM_ENGINES[typestr](**kwargs)  # type: ignore
        elif typestr in ILM_ENGINES_CLI:
            if not ILM_ENGINES_CLI[typestr].is_configured() and ILM_ENGINES_CLI[typestr].requires_configuration():
                raise ValueError(
                    f"Image/Language model type: {typestr} requires configuration. Please run `grazier configure {typestr}` first."
                )
            else:
                return ILM_ENGINES_CLI[typestr](**kwargs)  # type: ignore

        raise ValueError(f"Invalid Image/Language model type: {typestr}")

    @staticmethod
    def list_models() -> List[str]:
        return list(ILM_ENGINES_CLI.keys())


ILM_ENGINES: Dict[str, Type[ILMEngine]] = {}
ILM_ENGINES_CLI: Dict[str, Type[ILMEngine]] = {}

T = TypeVar("T")


def register_engine(cls: T) -> T:
    ILM_ENGINES[cls.name[0]] = cls  # type: ignore
    ILM_ENGINES_CLI[cls.name[1].lower()] = cls  # type: ignore
    return cls


# Imports for engine modules
from .blip_engine import *  # noqa: F403, E402
from .open_flamingo_engine import *  # noqa: F403, E402



from typing import Any, List, Optional

import torch
from PIL import Image
from transformers import BitsAndBytesConfig, Blip2ForConditionalGeneration, Blip2Processor

from grazier.engines.image import ILMEngine, register_engine
from grazier.utils.python import singleton
from grazier.utils.pytorch import select_device


class Blip2ILMEngine(ILMEngine):
    def __init__(self, model: str, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__(device=device)

        # Setup quantization
        if quantize:
            nf4_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_use_double_quant=True,
                bnb_4bit_compute_dtype=torch.float16,
            )
        else:
            nf4_config = None

        self._model = Blip2ForConditionalGeneration.from_pretrained(
            model,
            device_map="auto" if not self.device_specified else None,
            torch_dtype=torch.float16 if quantize else None,
            quantization_config=nf4_config,
        )
        if self.device_specified:
            self._model.to(self.device)
        self._processor = Blip2Processor.from_pretrained(
            model,
            # device=self.device if self.device_specified else None,
            device_map="auto" if not self.device_specified else None,
            torch_dtype=torch.float16 if quantize else None,
        )

    def call(self, image: Image, prompt: Optional[str] = None, n_completions: int = 1, **kwargs: Any) -> List[str]:
        image = image.convert("RGB")
        if prompt is None:
            prompt = "An photo of "

        inputs = self._processor(image, prompt, return_tensors="pt")

        # Convert the inputs to the correct device
        if self.device_specified:
            inputs = {k: v.to(self.device) for k, v in inputs.items() if isinstance(v, torch.Tensor)}

        outputs = self._model.generate(
            **inputs,
            max_new_tokens=kwargs.get("max_new_tokens", kwargs.pop("max_tokens", 128)),
            num_return_sequences=n_completions,
            do_sample=n_completions > 1,
            **kwargs,
        )

        outputs = self._processor.batch_decode(outputs, skip_special_tokens=True, clean_up_tokenization_spaces=False)

        return outputs


@register_engine
@singleton
class Blip2OPT27B(Blip2ILMEngine):
    name = ("Blip2 (OPT 2.7B)", "blip2-opt-2.7b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-opt-2.7b", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2OPT67B(Blip2ILMEngine):
    name = ("Blip2 (OPT 6.7B)", "blip2-opt-6.7b")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-opt-6.7b", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2OPT27BCoco(Blip2ILMEngine):
    name = ("Blip2 (OPT 2.7B, Coco Fine-tuned)", "blip2-opt-2.7b-coco")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-opt-2.7b-coco", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2OPT67BCoco(Blip2ILMEngine):
    name = ("Blip2 (OPT 6.7B, Coco Fine-tuned)", "blip2-opt-6.7b-coco")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-opt-6.7b-coco", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2FlanT5XL(Blip2ILMEngine):
    name = ("Blip2 (Flan-T5 XL)", "blip2-flan-t5-xl")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        # This forces a single device to be used, since there's a bug in Blip2 + Flan-T5, when the model is split across
        # multiple devices.
        device = select_device(device)
        super().__init__("Salesforce/blip2-flan-t5-xl", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2FlanT5XXL(Blip2ILMEngine):
    name = ("Blip2 (Flan-T5 XXL)", "blip2-flan-t5-xxl")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-flan-t5-xxl", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True


@register_engine
@singleton
class Blip2FlanT5XLCoco(Blip2ILMEngine):
    name = ("Blip2 (Flan-T5 XL, Coco Fine-tuned)", "blip2-flan-t5-xl-coco")

    def __init__(self, device: Optional[str] = None, quantize: bool = False) -> None:
        super().__init__("Salesforce/blip2-flan-t5-xl-coco", device=device, quantize=quantize)

    @staticmethod
    def is_configured() -> bool:
        return True



import os

from transformers.configuration_utils import cached_file, is_remote_url
from transformers.utils import CONFIG_NAME


def check_huggingface_model_files_are_local(pretrained_model_name_or_path: str, **kwargs):
    # Check to see if the model files are local
    cache_dir = kwargs.pop("cache_dir", None)
    force_download = kwargs.pop("force_download", False)
    resume_download = kwargs.pop("resume_download", False)
    proxies = kwargs.pop("proxies", None)
    use_auth_token = kwargs.pop("use_auth_token", None)
    revision = kwargs.pop("revision", None)
    subfolder = kwargs.pop("subfolder", "")
    from_pipeline = kwargs.pop("_from_pipeline", None)
    from_auto_class = kwargs.pop("_from_auto", False)
    commit_hash = kwargs.pop("_commit_hash", None)

    user_agent = {"file_type": "config", "from_auto_class": from_auto_class}
    if from_pipeline is not None:
        user_agent["using_pipeline"] = from_pipeline

    pretrained_model_name_or_path = str(pretrained_model_name_or_path)
    if os.path.isfile(os.path.join(subfolder, pretrained_model_name_or_path)):
        # Special case when pretrained_model_name_or_path is a local file
        resolved_config_file = pretrained_model_name_or_path
        return True
    elif is_remote_url(pretrained_model_name_or_path):
        return False
    else:
        configuration_file = kwargs.pop("_configuration_file", CONFIG_NAME)

        try:
            # Load from local folder or from cache or download from model Hub and cache
            resolved_config_file = cached_file(
                pretrained_model_name_or_path,
                configuration_file,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=True,
                use_auth_token=use_auth_token,
                user_agent=user_agent,
                revision=revision,
                subfolder=subfolder,
                _commit_hash=commit_hash,
            )
            if resolved_config_file is None:
                return False
            if os.path.isfile(resolved_config_file):
                return True
            return False
        except Exception:
            return False



import logging
from abc import abstractmethod
from typing import Any, List

import openai
from rich.prompt import Prompt

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.engines.llm.openai_engine import OpenAI
from grazier.utils.python import retry, singleton
from grazier.utils.secrets import get_secret, set_secret


def _setup_api_keys():
    # Setup openai api keys
    openai.api_key = get_secret("OPENAI_API_KEY", None)
    openai.organization = get_secret("OPENAI_API_ORG", None)


class OpenAIChatEngine(LLMChat):
    def __init__(self, model: str):
        super().__init__(device="api")

        _setup_api_keys()

        self._model = model

    @property
    @abstractmethod
    def cost_per_token(self) -> float:
        raise NotImplementedError()

    @retry(no_retry_on=(openai.error.AuthenticationError,))
    def _retry_call(self, *args: Any, **kwargs: Any) -> Any:
        return openai.ChatCompletion.create(*args, **kwargs)  # type: ignore

    def call(
        self,
        conversation: Conversation,
        n_completions: int = 1,
        **kwargs: Any,
    ) -> List[ConversationTurn]:
        # Construct the messages list from the conversation
        messages = []
        for turn in conversation.turns:
            messages.append(
                {
                    "role": "user"
                    if turn.speaker == Speaker.USER
                    else "system"
                    if turn.speaker == Speaker.SYSTEM
                    else "assistant"
                    if turn.speaker == Speaker.AI
                    else "user",
                    "content": turn.text,
                }
            )

        # Update temperature and max_tokens
        kwargs["temperature"] = kwargs.get("temperature", 0.9)
        kwargs["max_tokens"] = kwargs.get("max_tokens", 150)

        # Call the OpenAI API
        cp = self._retry_call(
            model=self._model,
            messages=messages,
            n=n_completions,
            **kwargs,
        )  # type: ignore
        OpenAI.USAGE += int(cp.usage.total_tokens) * self.cost_per_token

        return [
            ConversationTurn(
                text=i.message.content,
                speaker=Speaker.USER
                if i.message.role == "user"
                else Speaker.SYSTEM
                if i.message.role == "system"
                else Speaker.AI
                if i.message.role == "assistant"
                else Speaker.USER,
            )
            for i in cp.choices
        ]

    @staticmethod
    def is_configured() -> bool:
        _setup_api_keys()
        return openai.api_key is not None

    @staticmethod
    def configure():
        if OpenAIChatEngine.is_configured():
            reconfigure = Prompt.ask("OpenAI API key is already configured. Reconfigure?", choices=["y", "n"])
            if reconfigure == "n":
                logging.info("OpenAI API key already configured.")
                return

        have_openai_key = Prompt.ask("Do you have an OpenAI API key?", choices=["y", "n"])
        if have_openai_key == "n":
            Prompt.ask(
                "Please follow the instructions here [u cyan]https://www.howtogeek.com/885918/how-to-get-an-openai-api-key/[/u cyan] to obtain an API key.\nWhen you have the key, press enter to continue",
            )

        openai_key = Prompt.ask("Please enter your OpenAI API key (It will be saved to your platform's secrets store)")
        set_secret("OPENAI_API_KEY", openai_key.strip())

        # Save the openai api key to the secrets store
        logging.info("OpenAI API successfully configured.")


@register_engine
@singleton
class ChatGPT(OpenAIChatEngine):
    name = ("Chat GPT", "chat-gpt")
    cost_per_token = 0.002 / 1000

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("gpt-3.5-turbo")


@register_engine
@singleton
class GPT4(OpenAIChatEngine):
    name = ("GPT-4", "gpt4")
    cost_per_token = 0.03 / 1000

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("gpt-4")


@register_engine
@singleton
class GPT432K(OpenAIChatEngine):
    name = ("GPT-4 32K", "gpt4-32k")
    cost_per_token = 0.06 / 1000

    def __init__(self, **kwargs: Any) -> None:
        super().__init__("gpt-4-32k")



import logging
import os
from typing import Any, List

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import retry, singleton


@register_engine
@singleton
class BardEngine(LLMChat):
    name = ("Bard", "bard")

    def __init__(
        self,
    ) -> None:
        super().__init__(device="api")

        from Bard import Chatbot

        self._chatbot = Chatbot(
            os.environ.get("BARD__Secure_1PSID", None),
            os.environ.get("BARD__Secure_1PSIDTS", None),
        )

    @retry()
    def _ask(self, prompt: str) -> str:
        return self._chatbot.ask(prompt)["content"]

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        if len(conversation.turns) != 1:
            raise AssertionError("BARD conversations must have exactly one turn.")

        # Extract the last user turn from the conversation.
        user_turn = conversation.turns[-1]
        new_prompt = user_turn.text
        if user_turn.speaker in (Speaker.SYSTEM, Speaker.AI):
            raise AssertionError("The last turn in the conversation must be from the user.")

        outputs = []
        for _ in range(n_completions):
            logging.info("Sending prompt to BARD engine: %s", new_prompt)
            outputs.append(self._ask(new_prompt))

        return [ConversationTurn(speaker=Speaker.AI, text=output) for output in outputs]

    @staticmethod
    def is_configured() -> bool:
        return (
            os.environ.get("BARD__Secure_1PSID", None) is not None
            and os.environ.get("BARD__Secure_1PSIDTS", None) is not None
        )



from typing import Any, List, Optional

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.huggingface import check_huggingface_model_files_are_local
from grazier.utils.python import singleton
from grazier.utils.pytorch import select_device


class HuggingfaceChatEngine(LLMChat):
    def __init__(self, model: str, device: Optional[str] = None):
        device = select_device(device)
        self._tokenizer = AutoTokenizer.from_pretrained(model)
        self._model = AutoModelForCausalLM.from_pretrained(model)
        self._model.half().to(device)
        self._device = device

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        # Build the prompt
        chat = conversation.to_huggingface_chat()
        self._tokenizer.use_default_system_prompt = kwargs.get("use_default_system_prompt", False)
        inputs = self._tokenizer.apply_chat_template(chat, return_tensors="pt").to(self._device)

        # Get the params, and outputs
        _params = {
            "max_new_tokens": kwargs.get("max_new_tokens", kwargs.pop("max_tokens", 256)),
            "min_new_tokens": kwargs.get("min_new_tokens", 10),
            "num_return_sequences": n_completions,
            "temperature": kwargs.get("temperature", None),
            # "return_full_text": False,
        } | kwargs

        tokens = self._model.generate(
            inputs,
            do_sample=n_completions > 1,
            pad_token_id=self._tokenizer.pad_token_id,
            eos_token_id=self._tokenizer.eos_token_id,
            **_params,
        )

        last_turn_is_user = conversation.turns[-1].speaker == Speaker.USER
        outputs = [self._decode_tokens(t[inputs.shape[-1] :], last_turn_is_user) for t in tokens]  # type: ignore

        return [
            ConversationTurn(text=output, speaker=Speaker.AI if last_turn_is_user else Speaker.USER)
            for output in outputs
        ]

    def _decode_tokens(self, tokens: torch.Tensor, last_turn_is_user: bool) -> str:
        output = self._tokenizer.decode(tokens, skip_special_tokens=True).strip()
        return output

    @staticmethod
    def requires_configuration() -> bool:
        return False

    @classmethod
    def configure(cls, *args, **kwargs) -> None:
        pass


@register_engine
@singleton
class MistralInstruct7B(HuggingfaceChatEngine):
    name = ("Mistral Instruct (7B)", "mistral-instruct-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("mistralai/Mistral-7B-Instruct-v0.1", device=device)

    @staticmethod
    def is_configured() -> bool:
        return check_huggingface_model_files_are_local("mistralai/Mistral-7B-Instruct-v0.1")



import logging
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Type, TypeVar

from huggingface_hub import HfApi, ModelFilter

from grazier.engines.default import Engine
from grazier.utils.pytorch import select_device


class LLMEngine(Engine):
    def __init__(self, device: Optional[str] = None) -> None:
        self.device = select_device(device)
        self.device_specified = device is not None

    @property
    def prompt_prefix(self) -> str:
        return ""

    @property
    def prompt_suffix(self) -> str:
        return ""

    def __call__(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        prompt = self.prompt_prefix + prompt + self.prompt_suffix
        return self.call(prompt, n_completions, **kwargs)

    @abstractmethod
    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        raise NotImplementedError()

    def __repr__(
        self,
    ) -> str:
        return f"{self.__class__.__name__}({self.name[0]})"

    @staticmethod
    def from_string(typestr: str, **kwargs: Any) -> "LLMEngine":
        typestr = typestr.lower()

        if typestr in LM_ENGINES:
            if not LM_ENGINES[typestr].is_configured() and LM_ENGINES[typestr].requires_configuration():
                raise ValueError(
                    f"Language model type: {typestr} requires configuration. Please run `grazier configure {typestr}` first."
                )
            else:
                return LM_ENGINES[typestr](**kwargs)  # type: ignore
        elif typestr in LM_ENGINES_CLI:
            if not LM_ENGINES_CLI[typestr].is_configured() and LM_ENGINES_CLI[typestr].requires_configuration():
                raise ValueError(
                    f"Language model type: {typestr} requires configuration. Please run `grazier configure {typestr}` first."
                )
            else:
                return LM_ENGINES_CLI[typestr](**kwargs)  # type: ignore

        logging.info(f"Failed to find local LLM matching {typestr}. Fetching remote LLMs...")
        api = HfApi()
        models = list(api.list_models(filter=ModelFilter(model_name=typestr, task="text-generation")))
        if len(models) > 0:
            return HuggingFaceTextGenerationLMEngine.from_hub_model(typestr)(**kwargs)  # noqa: F405

        raise ValueError(f"Invalid language model type: {typestr}")

    @staticmethod
    def list_models() -> List[str]:
        return list(LM_ENGINES_CLI.keys())


LM_ENGINES: Dict[str, Type[LLMEngine]] = {}
LM_ENGINES_CLI: Dict[str, Type[LLMEngine]] = {}

T = TypeVar("T")


def register_engine(cls: T) -> T:
    LM_ENGINES[cls.name[0]] = cls  # type: ignore
    LM_ENGINES_CLI[cls.name[1].lower()] = cls  # type: ignore
    return cls


def register_chat_engine(cls: T) -> T:
    LM_ENGINES[cls.name[0]] = cls  # type: ignore
    if cls.name[1] not in LM_ENGINES_CLI:
        LM_ENGINES_CLI[cls.name[1].lower()] = cls  # type: ignore

    LM_ENGINES_CLI[f"{cls.name[1]}-chat".lower()] = cls  # type: ignore

    return cls


# Imports for engine modules
from grazier.engines.llm.ai21_engine import *  # noqa: F403, E402
from grazier.engines.llm.huggingface_engine import *  # noqa: F403, E402
from grazier.engines.llm.llama_engine import *  # noqa: F403, E402
from grazier.engines.llm.openai_engine import *  # noqa: F403, E402
from grazier.engines.llm.vertex_engine import *  # noqa: F403, E402



import os
from typing import Any, List

import ai21
from ai21.errors import (
    BadRequest,
    EmptyMandatoryListException,
    MissingApiKeyException,
    MissingInputException,
    NoSpecifiedRegionException,
    OnlyOneInputException,
    Unauthorized,
    UnprocessableEntity,
    UnsupportedDestinationException,
    UnsupportedInputException,
    WrongInputTypeException,
)

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.python import retry, singleton

ai21.api_key = os.environ.get("AI21_API_KEY", None)


class AI21CompletionLLMEngine(LLMEngine):
    def __init__(self, model: str) -> None:
        self._model = model
        super().__init__(device="api")

    @retry(
        no_retry_on=(
            BadRequest,
            Unauthorized,
            UnprocessableEntity,
            MissingInputException,
            UnsupportedInputException,
            UnsupportedDestinationException,
            OnlyOneInputException,
            WrongInputTypeException,
            EmptyMandatoryListException,
            MissingApiKeyException,
            NoSpecifiedRegionException,
        )
    )
    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        _default_params = {
            "numResults": n_completions,
            "maxTokens": 256,
            "temperature": 0.7,
            "stopSequences": ["##"],
        }

        response = ai21.Completion.execute(model=self._model, prompt=prompt, **(_default_params | kwargs))

        return [c["data"]["text"] for c in response["completions"]]

    @staticmethod
    def is_configured() -> bool:
        return ai21.api_key is not None


@register_engine
@singleton
class J2Light(AI21CompletionLLMEngine):
    name = ("J2 (Light)", "j2-light")

    def __init__(self) -> None:
        super().__init__("j2-light")


@register_engine
@singleton
class J2Mid(AI21CompletionLLMEngine):
    name = ("J2 (Mid)", "j2-mid")

    def __init__(self) -> None:
        super().__init__("j2-mid")


@register_engine
@singleton
class J2Ultra(AI21CompletionLLMEngine):
    name = ("J2 (Ultra)", "j2-ultra")

    def __init__(self) -> None:
        super().__init__("j2-ultra")
